import torch
import torch.nn as nn
import torch._utils
import torch.nn.functional as F
import os
import logging

logger = logging.getLogger(__name__)


class Net(nn.Module):
    def __init__(self, classes):
        super(Net, self).__init__()
        self.conv1 = nn.Conv2d(1, 60, (7, 5), padding='same')
        self.conv2 = nn.Conv2d(60, 40, (7, 5), padding='same')
        self.conv3 = nn.Conv2d(40, 60, (5, 5), padding='same')
        self.fc1 = nn.Linear(1440, 400)
        self.fc2 = nn.Linear(400, 200)
        self.fc3 = nn.Linear(200, classes)

    def forward(self, x):
        out = F.relu(self.conv1(x))
        out = F.dropout2d(out, 0.2, training=self.training)
        # print(out.shape)
        # out = F.embedding(out, max_norm=2.7)
        out = F.relu(self.conv2(out))
        out = F.dropout2d(out, 0.2, training=self.training)
        # out = F.embedding(out, max_norm=2.7)
        out = F.max_pool2d(out, 4)
        out = F.relu(self.conv3(out))
        out = F.dropout2d(out, 0.2, training=self.training)
        # out = F.embedding(out, max_norm=2.7)
        out = F.max_pool2d(out, (2, 4))
        # print(out.shape)
        out = torch.flatten(out, start_dim=1)
        # print(out.shape)
        out = F.relu(self.fc1(out))
        out = F.relu(self.fc2(out))
        out = self.fc3(out)
        return out

    def init_weights(self, pretrained='', ):
        logger.info('=> init weights from normal distribution')
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(
                    m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)
        if os.path.isfile(pretrained):
            pretrained_dict = torch.load(pretrained)
            # logger.info('=> loading pretrained model {}'.format(pretrained))
            model_dict = self.state_dict()
            pretrained_dict = {k: v for k, v in pretrained_dict.items()
                               if k in model_dict.keys()}
            for k, _ in pretrained_dict.items():
                logger.info(
                    '=> loading {} pretrained model {}'.format(k, pretrained))
            model_dict.update(pretrained_dict)
            self.load_state_dict(model_dict)


def get_net(is_train, classes):
    model = Net(classes=classes)
    if is_train:
        model.init_weights('')
    return model


if __name__ == '__main__':
    model = get_net(True, 3)
    input = torch.ones((2, 1, 99, 39))
    output = model(input)
    print(output.shape)
