import torch
import torch.nn as nn
import torch.utils.data as Data
from torch.utils.data import Dataset
from torchvision import transforms
from helper_code import *
from models import VAE
# import matplotlib.pyplot as plt
import scipy as sp
import scipy.signal

segTime = 0.5
step = 0.5
EPOCH = 20
BATCH_SIZE = 64
LR = 0.0005
BN_MOMENTUM = 0.1
NUM_WORKER = 4
is_resize = False
W = 2000
H = 1

def butterWorthFilter(x):
    b, a = sp.signal.butter(5, [0.0125, 0.75], 'bandpass')
    filtedData = sp.signal.filtfilt(b, a, x)
    return filtedData

def normalize(x):
    x1 = x - np.mean(x)
    # print(np.std(x))
    x2 = x1 / (np.std(x) + 1e-4)
    # if np.std(x) == 0:
    #     print('123')
    #     print(x)
    # print(x)
    # print(x2)
    return x2

def get_features_label(patient_files, data_folder):

    slices = list()
    slices_labels = list()
    count = 0
    for i in range(len(patient_files)):
    # for i in range(10):
        # i = i + 320
        # Load the current patient data and recordings.
        current_patient_data = load_patient_data(patient_files[i])
        current_recordings = load_recordings(data_folder, current_patient_data)
        freg = int(get_frequency(current_patient_data))
        location = get_locations(current_patient_data)
        # murmur_location = get_murmur_locations(current_patient_data)
        # label = get_label(current_patient_data)
        murmur_label = get_murmur(current_patient_data)
        outcome_label = get_outcome(current_patient_data)
        # print(outcome_label)

        # print(current_recordings)
        # print(current_recordings[0])
        # print(len(current_recordings))
        # print(location)
        # print(murmur_location)
        # print(label)

        # flag = list()
        # for j in range(len(current_recordings)):
        #     if murmur_label == 'Absent':
        #         flag.append(2)
        #         if outcome_label == 'Abnormal':
        #             flag.append(0)
        #         if outcome_label == 'Normal':
        #             flag.append(1)
        #     if murmur_label == 'Unknown':
        #         flag.append(1)
        #         if outcome_label == 'Abnormal':
        #             flag.append(0)
        #         if outcome_label == 'Normal':
        #             flag.append(1)
        #     if murmur_label == 'Present':
        #         flag.append(0)
        #         if outcome_label == 'Abnormal':
        #             flag.append(0)
        #         if outcome_label == 'Normal':
        #             flag.append(1)

        # if outcome_label == 'Abnormal':
        #     break
        if outcome_label == 'Normal' and murmur_label == 'Absent':
            for j in range(len(current_recordings)):
                x = butterWorthFilter(current_recordings[j])
                # x = current_recordings[j]
                n = np.ceil((len(x) - (segTime * freg) - (freg * step)) / (freg * step)) + 2
                for r in range(int(n)):
                    x_meg = x[int(r * freg * step):int(r * freg * step + segTime * freg)]
                    if (len(x_meg) == segTime * freg):
                        x_meg = normalize(x_meg)
                        slices.append(x_meg)
                        count = count + 1
                    if (len(x_meg) < segTime * freg):
                        x_meg = x[-int(segTime * freg):]
                        x_meg = normalize(x_meg)
                        slices.append(x_meg)
                        count = count + 1
                        break
    print(len(slices))

    x = torch.zeros((len(slices), 1, W))

    for j in range(len(slices)):

        Xdb = np.asarray(slices[j])
        # Xdb = MFCC.mfcc(slices[j], freg)
        # Xdb1 = np.concatenate((Xdb, MFCC.delta(Xdb, 2)), axis=1)
        # Xdb = np.concatenate((Xdb1, MFCC.delta(MFCC.delta(Xdb, 2), 2)), axis=1)

        # Xdb = Xdb.astype(np.float32)
        # Xdb = Xdb[:, :, np.newaxis]
        # Xdb = Xdb[np.newaxis, :]
        Xdb = Xdb[np.newaxis, :]

        # Xdb = normalize(Xdb)

        # print(Xdb.shape)
        # print(int(slices_labels[j]))
        x[j] = torch.from_numpy(Xdb)


    ratio = int(len(slices) / 10)
    start = 0
    end = ratio
    test_dat_x = x[start:end]
    train_dat_x_1 = x[0:start]
    train_dat_x_2 = x[end:len(slices)]
    train_dat_x = np.concatenate((train_dat_x_1, train_dat_x_2), axis=0)

    # joblib.dump(test_dat_x, open('testx.dat', "wb"))
    # joblib.dump(test_dat_y, open('testy.dat', "wb"))
    # joblib.dump(train_dat_x, open('trainx.dat', "wb"))
    # joblib.dump(train_dat_y, open('trainy.dat', "wb"))
    return train_dat_x, test_dat_x

# Train your model.
def pre_vae_model(data_folder, model_folder, verbose):
    test_loss = 1000
    # Find data files.
    if verbose >= 1:
        print('Finding data files...')

    # Find the patient data files.
    patient_files = find_patient_files(data_folder)
    num_patient_files = len(patient_files)

    if num_patient_files==0:
        raise Exception('No data was provided.')

    # Create a folder for the model if it does not already exist.
    os.makedirs(model_folder, exist_ok=True)


    # Extract the features and labels.
    if verbose >= 1:
        print('Extracting features and labels from the Challenge data...')


    # Extract features and label.
    features, test_features = get_features_label(patient_files, data_folder)
    print(features.shape)
    # get_features_label(patient_files, data_folder)

    # labels = joblib.load('trainy.dat')
    # print(labels.shape)
    # print(labels[0])
    # print(test_labels.shape)

    # print(loss_w)

    class spectrogram(Dataset):
        def __init__(self, features, test_features, train=True, is_resize=False, w=4000, h=1):
        # def __init__(self, train=True, is_resize=False, w=64, h=64):
            super().__init__()
            self.transform = transforms.ToTensor()
            self.is_resize = is_resize
            self.W = w
            self.H = h
            # if train:
            #     self.data_path = 'trainx.dat'
            #     self.label_path = 'trainy.dat'
            # else:
            #     self.data_path = 'testx.dat'
            #     self.label_path = 'testy.dat'
            #
            # self.data = joblib.load(self.data_path)
            # self.label = joblib.load(self.label_path)

            if train:
                self.data = features
            else:
                self.data = test_features

            # print(len(self.data), len(self.label))
            # print(self.data.shape, self.label.shape)

        def __getitem__(self, index):
                x = self.data[index]
                # x = self.transform(x)
                return x

        def __len__(self):
            return int(len(self.data))

    # Train the model.

    device = torch.device("cuda")
    torch_dataset = spectrogram(features, test_features, train=True, is_resize=is_resize, w=W, h=H)
    # torch_dataset = spectrogram(train=True, is_resize=is_resize, w=W, h=H)
    train_loader = Data.DataLoader(
        dataset=torch_dataset,
        batch_size=BATCH_SIZE,
        shuffle=True,
        num_workers=NUM_WORKER,
        drop_last=True
    )
    valid_torch_dataset = spectrogram(features, test_features, train=False, is_resize=is_resize, w=W, h=H)
    # valid_torch_dataset = spectrogram(train=False, is_resize=is_resize, w=W, h=H)
    valid_loader = Data.DataLoader(
        dataset=valid_torch_dataset,
        batch_size=BATCH_SIZE,
        shuffle=True,
        num_workers=NUM_WORKER,
        drop_last=True
    )
    print('数据加载完毕...')
    model = VAE.VanillaVAE(1, 128)
    model = model.to(device)
    # print(model)

    optimizer = torch.optim.Adam(model.parameters(), lr=LR)
    # scheduler = torch.optim.lr_scheduler.MultiStepLR(optimizer, milestones=[40, 50], gamma=0.1)
    # scheduler = torch.optim.lr_scheduler.MultiStepLR(optimizer, milestones=[50, 80], gamma=0.1)
    scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, T_max=50, eta_min=1e-6, )
    # loss_func = model.loss_function()
    # loss_func = nn.CrossEntropyLoss().to(device)  # 自带softmax
    # loss_func = nn.BCEWithLogitsLoss()
    # test_loss = 10000
    test_acc = 0

    for epoch in range(EPOCH):
        # count = 0
        model.train()
        for step, (x) in enumerate(train_loader):
            # print(step)
            b_x = torch.tensor(x).to(device)

            # output = torch.tensor(model(b_x)).to(device)
            output = model(b_x)
            # print(output[0].shape)
            loss = model.loss_function(*output, M_N=1.0)['loss']
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            if step % 50 == 0:
                print('Epoch: ', epoch, '| train loss: %.4f' % loss,
                      '| Reconstruction_Loss: %.4f' % model.loss_function(*output)['Reconstruction_Loss'],
                      '| KLD: %.4f' % model.loss_function(*output)['KLD'])
        scheduler.step()


        # 测试
        valid_loss = 0
        count = 0
        with torch.no_grad():
            model.eval()
            for step, (x) in enumerate(valid_loader):
                valid_x = x.to(device)
                output = model(valid_x)
                valid_loss = valid_loss + model.loss_function(*output)['loss']
                count = count + 1
            # Save the model.
            if test_loss > valid_loss:
                test_loss = valid_loss
                torch.save(model.state_dict(), os.path.join(model_folder, 'pre_vae2.pkl'))
                print("准确模型更新")
            print('Epoch: ', epoch, '| test loss: %.4f' % (valid_loss / count))

        # if (epoch + 1) % 1 == 0:
        #     print(valid_x[0].cpu().detach().numpy()[0].shape)
        #     plt.subplot(211)
        #     plt.plot(np.arange(0, 2000), normalize(valid_x[0].cpu().detach().numpy()[0]))
        #     plt.subplot(212)
        #     # Pxx, fre = plt.psd(s, 512, 1 / dt)
        #     plt.plot(np.arange(0, 2000), model.generate(valid_x)[0].cpu().detach().numpy()[0])
        #     plt.show()
        if (epoch + 1) % 20 == 0:
               torch.save(model.state_dict(), os.path.join(model_folder, str(epoch + 1) + 'pre_vae2.pkl'))
               print("迭代模型更新")

if __name__ == "__main__":
    data_folder = '/home/tsz/Cinc2022_match/training_data'
    model_folder = '/home/tsz/Cinc2022_match/model'
    verbose = 3
    train_challenge_model(data_folder, model_folder, verbose)
