#!/usr/bin/env python

# Edit this script to add your team's code. Some functions are *required*, but you can edit most parts of the required functions,
# change or remove non-required functions, and add your own functions.

################################################################################
#
# Import libraries and functions. You can change or remove them.
#
################################################################################

from helper_code import *
import numpy as np, scipy as sp, scipy.stats, os, sys, joblib
import scipy.signal
from models import VAE
import torch
import torch.nn as nn
import torch.utils.data as Data
from torch.utils.data import Dataset
from torchvision import transforms
from models.match_net import get_net
from pre_train_vae import pre_vae_model
# import run_model
# import evaluate_model
################################################################################
#
# Required functions. Edit these functions to add your code, but do not change the arguments.
#
################################################################################
# hyper-parameter
segTime = 0.5
step = 0.5
EPOCH = 40
BATCH_SIZE = 128
LR = 0.0001
BN_MOMENTUM = 0.1
NUM_WORKER = 4
is_resize = False
W = 2000
H = 1


def get_loss_w1(l):
    n_0 = 0
    n_1 = 0
    n_2 = 0
    n = len(l)
    for i in range(n):
        if l[i] == 0:
            n_0 = n_0 + 1
        if l[i] == 1:
            n_1 = n_1 + 1
        if l[i] == 2:
            n_2 = n_2 + 1

    return torch.tensor([1 - n_0 / n, 1 - n_1 / n, 1 - n_2 / n])


def get_loss_w2(l):
    n_0 = 0
    n_1 = 0
    n = len(l)
    for i in range(n):
        if l[i] == 0:
            n_0 = n_0 + 1
        if l[i] == 1:
            n_1 = n_1 + 1


def compute_probabilities(pro_list, classes, out_pro_list, out_classes):
    global out_list, out_pro, mur_list, mur_pro
    flag = None
    labels = np.zeros(len(classes), dtype=np.int_)
    for i in range(len(pro_list)):
        idx = np.argmax(pro_list[i].cpu().numpy().tolist())
        labels[idx] = labels[idx] + 1
        print(labels)
        if labels[0] == 1:
            mur_list = [1, 0, 0]
            mur_pro = pro_list[i].cpu().numpy().tolist()
        if labels[1] == 1:
            flag = i
    if labels[2] == len(pro_list):
        mur_list = [0, 0, 1]
        mur_pro = pro_list[0].cpu().numpy().tolist()
    if labels[2] < len(pro_list) and labels[0] == 0:
        mur_list = [0, 1, 0]
        mur_pro = pro_list[flag].cpu().numpy().tolist()

    out_labels = np.zeros(len(out_classes), dtype=np.int_)
    for i in range(len(out_pro_list)):
        idx = np.argmax(out_pro_list[i].cpu().numpy().tolist())
        out_labels[idx] = out_labels[idx] + 1
        if out_labels[0] == 1:
            out_list = [1, 0]
            out_pro = out_pro_list[i].cpu().numpy().tolist()
    if out_labels[1] == len(out_pro_list):
        out_list = [0, 1]
        out_pro = out_pro_list[0].cpu().numpy().tolist()
    mur_list.extend(out_list)
    mur_pro.extend(out_pro)
    return mur_list, mur_pro


def normalize(x):
    x1 = x - np.mean(x)
    x2 = x1 / (np.std(x) + 1e-4)
    return x2


def get_murmur_locations(data):
    murmur_location = list()
    for l in data.split('\n'):
        if l.startswith('#Murmur locations: '):
            try:
                murmur_location = l.split(': ')[1].strip()
                murmur_location = murmur_location.split('+')
            except:
                pass
    return murmur_location


def butterWorthFilter(x):
    b, a = sp.signal.butter(5, [0.0125, 0.75], 'bandpass')
    filtedData = sp.signal.filtfilt(b, a, x)
    return filtedData


# Train your model.
def train_challenge_model(data_folder, model_folder, verbose):
    # Find data files.
    pre_vae_model(data_folder, model_folder, verbose)
    if verbose >= 1:
        print('Finding data files...')

    # Find the patient data files.
    patient_files = find_patient_files(data_folder)
    num_patient_files = len(patient_files)

    if num_patient_files == 0:
        raise Exception('No data was provided.')

    # Create a folder for the model if it does not already exist.
    os.makedirs(model_folder, exist_ok=True)

    # Extract the features and labels.
    if verbose >= 1:
        print('Extracting features and labels from the Challenge data...')

    # Extract features and label.
    features, mur_labels, out_labels, test_features, test_mur_labels, test_out_labels = get_features_label(
        patient_files, data_folder)
    # print(features[0])
    # print(mur_labels[0])
    # print(out_labels[0])
    # print(test_features.shape)
    # print(test_mur_labels.shape)
    # print(test_out_labels.shape)
    # get_features_label(patient_files, data_folder)

    # labels = joblib.load('trainy.dat')
    # print(labels.shape)
    # print(labels[0])
    # print(test_labels.shape)

    loss_w1 = get_loss_w1(mur_labels)
    loss_w2 = get_loss_w2(out_labels)

    # print(loss_w)

    class spectrogram(Dataset):
        def __init__(self, features, mur_labels, out_labels, test_features, test_mur_labels, test_out_labels,
                     train=True):
            # def __init__(self, train=True, is_resize=False, w=64, h=64):
            super().__init__()
            self.transform = transforms.ToTensor()

            if train:
                self.data = features
                self.mur_label = mur_labels
                self.out_label = out_labels
            else:
                self.data = test_features
                self.mur_label = test_mur_labels
                self.out_label = test_out_labels

            # print(len(self.data), len(self.label))
            # print(self.data.shape, self.label.shape)

        def __getitem__(self, index):

            train_data = np.asarray(self.data[index])
            train_data = train_data[np.newaxis, :]
            mur_y = self.mur_label[index]
            out_y = self.out_label[index]
            return train_data, mur_y, out_y

        def __len__(self):
            return int(len(self.data))

    # Train the model.

    device = torch.device("cuda")
    torch_dataset = spectrogram(features, mur_labels, out_labels, test_features, test_mur_labels, test_out_labels,
                                train=True)
    # torch_dataset = spectrogram(train=True, is_resize=is_resize, w=W, h=H)
    train_loader = Data.DataLoader(
        dataset=torch_dataset,
        batch_size=BATCH_SIZE,
        shuffle=True,
        num_workers=NUM_WORKER,
        drop_last=True
    )
    valid_torch_dataset = spectrogram(features, mur_labels, out_labels, test_features, test_mur_labels, test_out_labels,
                                      train=False)
    # valid_torch_dataset = spectrogram(train=False, is_resize=is_resize, w=W, h=H)
    valid_loader = Data.DataLoader(
        dataset=valid_torch_dataset,
        batch_size=BATCH_SIZE,
        shuffle=True,
        num_workers=NUM_WORKER,
        drop_last=True
    )
    print('数据加载完毕...')
    pre_model = VAE.VanillaVAE(1, 128)
    pre_model.load_state_dict(torch.load(os.path.join(model_folder, 'pre_vae2.pkl')))
    pre_model = pre_model.to(device)
    model = get_net(True, [3, 2]).to(device)
    # print(model)
    optimizer = torch.optim.Adam(model.parameters(), lr=LR)
    # scheduler = torch.optim.lr_scheduler.MultiStepLR(optimizer, milestones=[40, 70], gamma=0.1)
    scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, T_max=100, eta_min=1e-6, )
    loss_func1 = nn.CrossEntropyLoss(weight=loss_w1).to(device)  # 自带softmax
    loss_func2 = nn.CrossEntropyLoss(weight=loss_w2).to(device)  # 自带softmax
    # loss_func = nn.BCEWithLogitsLoss()
    # test_loss = 10000
    test_acc = 0

    for epoch in range(EPOCH):
        # count = 0
        model.train()
        # pre_model.eval()
        for step, (x, y1, y2) in enumerate(train_loader):
            # print(step)
            b_x = x
            b_x = b_x.type(torch.FloatTensor).to(device)
            b_y1 = y1
            b_y1 = b_y1.type(torch.FloatTensor).to(device)
            b_y2 = y2
            b_y2 = b_y2.type(torch.FloatTensor).to(device)

            mu, log_var = pre_model.encode(b_x)
            z = pre_model.reparameterize(mu, log_var).to(device)
            z = z[:, np.newaxis, :]

            output = model(z.detach())
            # output = model(b_x)

            loss1 = loss_func1(output[0], b_y1.long())
            loss2 = loss_func2(output[1], b_y2.long())
            loss = loss1 + loss2
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            if step % 50 == 0:
                pred_y1 = torch.max(output[0], 1)[1].data.squeeze()
                pred_y2 = torch.max(output[1], 1)[1].data.squeeze()
                print(optimizer.param_groups[0]['lr'])
                accuracy1 = (pred_y1 == b_y1).sum().item() / b_y1.size(0)
                accuracy2 = (pred_y2 == b_y2).sum().item() / b_y2.size(0)
                print('Epoch: ', epoch, '| train loss: %.4f' % loss.item(), '| train mur_accuracy: %.2f' % accuracy1,
                      '| train out_accuracy: %.2f' % accuracy2)
        scheduler.step()
        # 测试
        valid_loss = 0
        valid_acc = 0
        mur_valid_acc = 0
        out_valid_acc = 0
        count = 0
        with torch.no_grad():
            model.eval()
            # pre_model.eval()
            for step, (x, y1, y2) in enumerate(valid_loader):
                valid_x = x
                valid_x = valid_x.type(torch.FloatTensor).to(device)
                valid_y1 = y1
                valid_y1 = valid_y1.type(torch.FloatTensor).to(device)
                valid_y2 = y2
                valid_y2 = valid_y2.type(torch.FloatTensor).to(device)

                mu, log_var = pre_model.encode(valid_x)
                z = pre_model.reparameterize(mu, log_var).to(device)
                z = z[:, np.newaxis, :]
                output = model(z)
                # output = model(valid_x)

                pred_y1 = torch.max(output[0], 1)[1].data.squeeze()
                pred_y2 = torch.max(output[1], 1)[1].data.squeeze()
                valid_loss = valid_loss + loss_func1(output[0], valid_y1.long()) + loss_func2(output[1],
                                                                                              valid_y2.long())
                valid_acc = valid_acc + ((pred_y1 == valid_y1).sum().item() / valid_y1.size(0) +
                                         (pred_y2 == valid_y2).sum().item() / valid_y2.size(0)) / 2
                mur_valid_acc = mur_valid_acc + (pred_y1 == valid_y1).sum().item() / valid_y1.size(0)
                out_valid_acc = out_valid_acc + (pred_y2 == valid_y2).sum().item() / valid_y2.size(0)

                count = count + 1
            # Save the model.
            if valid_acc > test_acc:
                test_acc = valid_acc
                # test_loss = valid_loss
                torch.save(model.state_dict(), os.path.join(model_folder, 'work.pkl'))
                print("准确模型更新")
            print('Epoch: ', epoch, '| test loss: %.4f' % (valid_loss / count),
                  '| test accuracy: %.2f' % (valid_acc / count))
            print('| test mur_accuracy: %.2f' % (mur_valid_acc / count),
                  '| test out_accuracy: %.2f' % (out_valid_acc / count))

        if (epoch + 1) % 20 == 0:
            torch.save(model.state_dict(), os.path.join(model_folder, str(epoch + 1) + 'work.pkl'))
            print("迭代模型更新")


# Load your trained model. This function is *required*. You should edit this function to add your code, but do *not* change the
# arguments of this function.
def load_challenge_model(model_folder, verbose):
    filename = os.path.join(model_folder, 'work.pkl')
    model = get_net(True, [3, 2])
    model.load_state_dict(torch.load(filename))
    return model


# Run your trained model. This function is *required*. You should edit this function to add your code, but do *not* change the
# arguments of this function.
def run_challenge_model(model, data, recordings, verbose):
    device = torch.device("cuda:0")
    mur_classes = ['Present', 'Unknown', 'Absent']
    out_classes = ['Abnormal', 'Normal']
    classes = ['Present', 'Unknown', 'Absent', 'Abnormal', 'Normal']
    # probabilities = torch.zeros([3], dtype=torch.float).to(device)
    mur_probabilities = list()
    out_probabilities = list()
    # Load features.
    # patient_files = find_patient_files(data)
    freg = int(get_frequency(data))
    location = get_locations(data)
    murmur_location = get_murmur_locations(data)
    # label = get_label(data)

    for j in range(len(recordings)):
        # if label == 'Absent':
        #     flag = 2
        # if label == 'Unknown':
        #     flag = 1
        # if label == 'Present':
        #     flag = 0
        #     if location[j] not in murmur_location:
        #         flag = 2
        # x = butterWorthFilter(recordings[j])
        x = recordings[j]
        n = np.ceil((len(x) - (segTime * freg) - (freg * step)) / (freg * step)) + 2
        # print(int(n*len(recordings)))
        all_x = torch.zeros([int(n), 1, W], dtype=torch.float)
        for k in range(int(n)):
            x_meg = x[int(k * freg * step):int(k * freg * step + segTime * freg)]
            if (len(x_meg) < segTime * freg):
                x_meg = x[-int(segTime * freg):]
            x_meg = normalize(x_meg)
            x_meg = torch.Tensor(x_meg)

            x_meg = x_meg[np.newaxis, :]

            # Xdb = MFCC.mfcc(x_meg, freg)
            # Xdb1 = np.concatenate((Xdb, MFCC.delta(Xdb, 2)), axis=1)
            # Xdb = np.concatenate((Xdb1, MFCC.delta(MFCC.delta(Xdb, 2), 2)), axis=1)
            # Xdb = Xdb.astype(np.float32)
            # Xdb = Xdb[:, :, np.newaxis]
            # Xdb = Xdb[np.newaxis, :]
            # Xdb = normalize(Xdb)
            # print(Xdb.shape)

            # Xdb = trans_data(Xdb, is_resize)
            all_x[k] = x_meg
        # print(all_x.shape)
        # 一条记录集
        # all_x = torch.Tensor(all_x)
        model = model.to(device)
        model.eval()
        with torch.no_grad():
            all_x = all_x.to(device)
            # all_y = all_y.to(device)

            # label = label.to(device)
            # print(label.item())
            output = model(all_x)
            mur_output = torch.nn.Softmax(dim=1)(output[0])
            out_output = torch.nn.Softmax(dim=1)(output[1])
            # print(output)

            # probabilities = probabilities + torch.mean(output, dim=0, keepdim=False)
            mur_probabilities.append(torch.mean(mur_output, dim=0, keepdim=False))
            out_probabilities.append(torch.mean(out_output, dim=0, keepdim=False))

    labels, probabilities = compute_probabilities(mur_probabilities, mur_classes, out_probabilities, out_classes)

    return classes, labels, probabilities


################################################################################
#
# Optional functions. You can change or remove these functions and/or add new functions.
#
################################################################################


# Extract features from the data.
def get_features_label(patient_files, data_folder):
    slices = list()
    slices_labels = list()
    slices_labels2 = list()
    count = 0
    flag1 = None
    flag2 = None
    for i in range(len(patient_files)):
        # for i in range(10):
        # i = i + 320
        # Load the current patient data and recordings.
        current_patient_data = load_patient_data(patient_files[i])
        current_recordings = load_recordings(data_folder, current_patient_data)
        freg = int(get_frequency(current_patient_data))
        location = get_locations(current_patient_data)
        murmur_label = get_murmur(current_patient_data)
        murmur_location = get_murmur_locations(current_patient_data)
        outcome_label = get_outcome(current_patient_data)
        # print(current_recordings)
        # print(current_recordings[0])
        # print(len(current_recordings))
        # print(location)
        # print(murmur_location)
        # print(label)

        for j in range(len(current_recordings)):
            if murmur_label == 'Absent':
                flag1 = 2
                if outcome_label == 'Abnormal':
                    flag2 = 0
                if outcome_label == 'Normal':
                    flag2 = 1
            if murmur_label == 'Unknown':
                flag1 = 1
                if outcome_label == 'Abnormal':
                    flag2 = 0
                if outcome_label == 'Normal':
                    flag2 = 1
            if murmur_label == 'Present':
                flag1 = 0
                if location[j] not in murmur_location:
                    flag1 = 2
                if outcome_label == 'Abnormal':
                    flag2 = 0
                if outcome_label == 'Normal':
                    flag2 = 1

            # for j in range(len(current_recordings)):
            #     if label == 'Absent':
            #         flag = 2
            #     if label == 'Unknown':
            #         flag = 1
            #     if label == 'Present':
            #         flag = 0
            #         if location[j] not in murmur_location:
            #             flag = 2
            # x = butterWorthFilter(current_recordings[j])
            x = current_recordings[j]
            n = np.ceil((len(x) - (segTime * freg) - (freg * step)) / (freg * step)) + 2
            for r in range(int(n)):
                x_meg = x[int(r * freg * step):int(r * freg * step + segTime * freg)]
                x_meg = normalize(x_meg)
                if (len(x_meg) == segTime * freg):
                    slices.append(x_meg)
                    slices_labels.append(flag1)
                    slices_labels2.append(flag2)
                    count = count + 1
                if (len(x_meg) < segTime * freg):
                    x_meg = x[-int(segTime * freg):]
                    x_meg = normalize(x_meg)
                    slices.append(x_meg)
                    slices_labels.append(flag1)
                    slices_labels2.append(flag2)
                    count = count + 1
                    break

    # print(len(slices))
    # print(len(slices_labels))
    # print(count)
    # print(len(slices[0]))
    # x = np.zeros((len(slices), W, H, 1), np.float32) # MFCC
    # x = np.zeros((len(rows), 129, 33, 1), np.float32)  # STFT
    # y = np.zeros((len(slices), 1), int)

    x = np.asarray(slices)
    y = np.asarray(slices_labels)
    y2 = np.asarray(slices_labels2)

    # for j in range(len(slices)):
    #
    #     Xdb = STFT.trans_stft(slices[j], freg)
    #     # Xdb = MFCC.mfcc(slices[j], freg)
    #     # Xdb1 = np.concatenate((Xdb, MFCC.delta(Xdb, 2)), axis=1)
    #     # Xdb = np.concatenate((Xdb1, MFCC.delta(MFCC.delta(Xdb, 2), 2)), axis=1)
    #
    #     Xdb = Xdb.astype(np.float32)
    #     Xdb = Xdb[:, :, np.newaxis]
    #     # Xdb = Xdb[np.newaxis, :]
    #     Xdb = Xdb[np.newaxis, :]
    #
    #     Xdb = normalize(Xdb)
    #
    #     # print(Xdb.shape)
    #     # print(int(slices_labels[j]))
    #     x[j] = Xdb
    #     # y[num[j]] = label
    #     y[j] = int(slices_labels[j])

    ratio = int(len(slices) / 10)
    start = 0
    end = ratio
    test_dat_x = x[start:end]
    test_dat_y = y[start:end]
    test_dat_y2 = y2[start:end]

    train_dat_x_1 = x[0:start]
    train_dat_x_2 = x[end:len(slices)]
    train_dat_x = np.concatenate((train_dat_x_1, train_dat_x_2), axis=0)

    train_dat_y_1 = y[0:start]
    train_dat_y_2 = y[end:len(slices)]
    train_dat_y = np.concatenate((train_dat_y_1, train_dat_y_2), axis=0)

    train_dat_y2_1 = y2[0:start]
    train_dat_y2_2 = y2[end:len(slices)]
    train_dat_y2 = np.concatenate((train_dat_y2_1, train_dat_y2_2), axis=0)

    return train_dat_x, train_dat_y, train_dat_y2, test_dat_x, test_dat_y, test_dat_y2


# if __name__ == '__main__':
    # data = '/home/tsz/Cinc2022_match/training_data'
    # mod = '/home/tsz/Cinc2022_match/model'
    # # train_challenge_model(data_folder=data, model_folder=mod, verbose=3)
    # test_data = '/home/tsz/Cinc2022_match/test_data'
    # test_outputs = '/home/tsz/Cinc2022_match/test_outputs'
    # # run_model.run_model(mod, test_data, test_outputs, False, 3)
    # murmur_scores, outcome_scores = evaluate_model.evaluate_model(test_data, test_outputs)
    # print(murmur_scores, outcome_scores)
